(* main.sml
 *
 * COPYRIGHT (c) 1995 AT&T Bell Laboratories.
 *
 * A tool for extracting signatures from ML-Doc documents.
 *)

structure Main : sig

    val main : string * string list -> OS.Process.status

  end = struct

    structure C = MLDocConfig
    structure E = MLDocElem
    structure P = MLDocParser
    structure M = P.Markup
    structure Ty = MLDocType

    structure SS = Substring
    structure F = Format

    val sdata = EntitiesToSML.transEntity

    val transData = M.transData {
	    escape = String.str,
	    sdata = sdata,
	    special = fn _ => NONE
	  }

    fun getCopyrights iter = let
	  fun find (cpys, iter) = (case (M.next iter)
		 of NONE => NONE
		  | (SOME(M.ELEM{elem=E.SECTION _, ...}, _)) =>
		      SOME(rev cpys, iter)
		  | (SOME(M.ELEM{elem=E.COPYRIGHT cpy, ...}, iter')) =>
		      find (cpy::cpys, iter')
		  | (SOME(_, iter')) => find (cpys, iter')
		(* end case *))
	  in
	    find ([], iter)
	  end

  (* scan the markup tree for the next SIGBODY *)
    fun findSigBody iter = let
	  fun find NONE = NONE
	    | find (SOME(M.ELEM{
		  elem=E.SIGBODY{sigid=SOME id, file=SOME f}, body, ...
		}, iter')) =
		  SOME({sigId=id, file=transData f, sigBody=body}, M.right iter')
	    | find (SOME(M.ELEM{
		  elem=E.SIGNATURE{sigid=id, ...}, body = (M.ELEM{
		    elem=E.SIGBODY{file=SOME f,...}, body, ...}
                    )::_, ...
		}, iter')) =
		  SOME({sigId=id, file=transData f, sigBody=body}, M.right iter')
	    | find (SOME(_, iter')) = find(M.next iter')
	  in
	    find(M.next iter)
	  end

  (* process a line of code from a signature body.  The code is passed in
   * as a reversed list of chunks. Eventually, we want to be able to
   * set the indentation for nested structures.
   *)
    local
      val stripr = SS.dropr (Char.contains " \t")
      val stripl = SS.dropl Char.isSpace
      val indent = "    "
    in
    fun doCodeLn [] = "\n"
      | doCodeLn [line] = indent ^ SS.string (stripl (stripr (SS.all line)))
      | doCodeLn (last::rest) = let
	  val last' = SS.string (stripr (SS.all last))
	  val (first :: rest) = rev (last' :: rest)
	  val first' = SS.string (stripl (SS.all first))
	  in
	    String.concat(indent :: first' :: rest)
	  end
    end (* local *)

    fun generateFile {file, cpys, sigid, specs} = let
	  val outS = TextIO.openOut file
	  fun pr s = TextIO.output(outS, s)
	  fun prf (fmt, items) = pr(F.format fmt items)
	  fun dataToStr data = SS.string(
		SS.dropl Char.isSpace (
		  SS.dropr Char.isSpace (SS.all (transData data))))
	  fun fmtData data = F.STR(dataToStr data)
	  fun prIndent 0 = ()
	    | prIndent 1 = pr "  "
	    | prIndent 2 = pr "    "
	    | prIndent 3 = pr "      "
	    | prIndent k = (pr "      "; prIndent(k-3))
	  fun prCopyright {owner, year} = prf(
		" * COPYRIGHT (c) %d %s.\n",
		[F.INT year, fmtData owner]);
	  fun tagIdToStr (MLSpec.ID s) = dataToStr s
	    | tagIdToStr (MLSpec.TAG m) = let
		fun markupToStr (MLDocMarkup.DATA id) = dataToStr id
		  | markupToStr (
			MLDocMarkup.ELEM{body=[MLDocMarkup.DATA id], ...}
		      ) = dataToStr id
		  | markupToStr m = Error.bogusMarkup ("tagged-id", m)
		in
		  markupToStr m
		end
	  fun fmtTy ty = let
		fun fmtCon (id, txt) = tagIdToStr id :: txt
		fun fmtList (left, right, fmtItem) = let
		      fun fmt ([], txt) = txt
			| fmt ([x], txt) = fmtItem(x, txt)
			| fmt (x::r, txt) = fmtItem(x, ", "::fmt(r, txt))
		      in
			fn (l, txt) => left :: fmt(l, right::txt)
		      end
		fun fmt (Ty.VarTy s, txt) = s :: txt
		  | fmt (Ty.BaseTy([], con), txt) = fmtCon (con, txt)
		  | fmt (Ty.BaseTy([ty], con), txt) =
		      fmt (ty, " " :: fmtCon (con, txt))
		  | fmt (Ty.BaseTy(tys, con), txt) =
		      fmtTys (tys, " " :: fmtCon (con, txt))
		  | fmt (Ty.FnTy(ty1, ty2), txt) =
		      fmt (ty1, " -> " :: fmt(ty2, txt))
		  | fmt (Ty.TupleTy tys, txt) = fmtTys (tys, txt)
		  | fmt (Ty.RecordTy ltys, txt) = let
		      fun fmtItem ((l, ty), txt) = l :: " : " :: fmt(ty, txt)
		      in
			fmtList ("{", "}", fmtItem) (ltys, txt)
		      end
		  | fmt (Ty.ParenTy ty, txt) = "(" :: fmt(ty, ")" :: txt)
		and fmtTys (tys, txt) = fmtList ("(", ")", fmt) (tys, txt)
		in
		  F.STR(concat(fmt(ty, [])))
		end
	  fun prSpecGrps (indent, []) = ()
	    | prSpecGrps (indent, [specGrp]) = (
		pr "\n"; prSpecs(indent, [#1 specGrp]); pr "\n")
	    | prSpecGrps (indent, specGrp::rest) = (
		pr "\n"; prSpecs(indent, [#1 specGrp]);
		prSpecGrps (indent, rest))
	  and prSpecs (indent, []) = ()
	    | prSpecs (indent, spec::r) = let
		fun fmtTyParams NONE = F.STR ""
		  | fmtTyParams (SOME params) =
		      F.STR(F.format "%s " [fmtData params])
		fun prWhereTy (MLSpec.WHEREty{params, id, def}) = (
			prIndent (indent+1);
			prf("where type %s%s = %s\n", [
			    fmtTyParams params, fmtData id, fmtTy def
			  ]))
		fun prConsList (compact, cons) = let
		      val sep = if compact
			    then (fn () => pr " ")
			    else (fn () => (pr "\n"; prIndent(indent+1)))
		      fun prCons (id, NONE, _) =
			    prf ("%s", [fmtData id])
			| prCons (id, SOME ty, _) =
			    prf("%s of %s", [fmtData id, fmtTy ty])
		      fun prConsList [c] = prCons c
			| prConsList (c::r) = (
			    prCons c; sep(); pr "| "; prConsList r)
		      in
			sep(); pr "= "; prConsList cons; pr "\n"
		      end
		fun fmtMarkupId id = F.STR(tagIdToStr id)
                fun prINCLspec (id, wtys) = (
			prIndent indent;
			prf("include %s\n", [fmtMarkupId id]);
			List.app prWhereTy wtys)
		fun prEXNspec (id, NONE) = (
			prIndent indent;
			prf("exception %s\n", [fmtData id]))
		  | prEXNspec (id, SOME ty) = (
			prIndent indent;
			prf("exception %s of %s\n", [fmtData id, fmtTy ty]))
		fun prTYspec {eq, params, id, def} = (
			prIndent indent;
			prf("%s %s%s", [
			    if eq then F.STR "eqtype " else F.STR "type ",
			    fmtTyParams params, fmtData id
			  ]);
			case def
			 of NONE => pr "\n"
			  | (SOME ty) => prf(" = %s\n", [fmtTy ty])
			(* end case *))
		fun prDTDEFspec {id, def} = (
			prIndent indent;
			prf("datatype %s = datatype %s\n", [
			    fmtData id, fmtMarkupId def
			  ]))
		fun prVALspec (id, ty, _) = (
			prIndent indent;
			prf("val %s : %s\n", [fmtData id, fmtTy ty]))
		in
		  case spec
		   of (MLSpec.BRspec false) => ()
		    | (MLSpec.BRspec true) => pr "\n"
		    | (MLSpec.INCLspec incls) => app prINCLspec incls
		    | (MLSpec.STRspec(strId, sigId, wtys)) => (
			prIndent indent;
			prf("structure %s : %s\n",
			  [F.STR strId, fmtMarkupId sigId]);
			List.app prWhereTy wtys)
		    | (MLSpec.STRSIGspec(strId, specs, wtys)) => (
			prIndent indent;
			prf("structure %s : sig\n", [fmtData strId]);
			prSpecGrps (indent+2, specs);
			prIndent(indent+1); pr "end\n";
			List.app prWhereTy wtys)
		    | (MLSpec.SHARINGspec l) => let
			fun prIds [id] = prf("%s\n", [fmtData id])
			  | prIds (id::r) = (prf("%s = ", [fmtData id]); prIds r)
			fun prSharing (MLSpec.STRshare ids) = (
			      prIndent(indent+1); pr "sharing "; prIds ids)
			  | prSharing (MLSpec.TYshare ids) = (
			      prIndent(indent+1); pr "sharing type "; prIds ids)
			in
		          List.app prSharing l
			end
		    | (MLSpec.EXNspec exns) => app prEXNspec exns
		    | (MLSpec.TYspec tys) => app prTYspec tys
		    | (MLSpec.DTspec(dt::rest)) => let
			fun prDT kw {compact, params, id, cons} = (
			      prIndent indent;
			      prf("%s %s%s",
				[F.STR kw, fmtTyParams params, fmtData id]);
			      prConsList (compact, cons))
			in
			  prDT "datatype" dt;
			  List.app (prDT "and") rest
			end
		    | (MLSpec.DTDEFspec defs) => app prDTDEFspec defs
		    | (MLSpec.VALspec vals) => app prVALspec vals
		  (* end case *);
		  prSpecs (indent, r)
		end
	  in
	    prf ("\
		\(* %s\n\
		\ *\n\
	      \", [F.STR(OS.Path.file file)]);
	    app prCopyright cpys;
	    prf ("\
		\ *)\n\
		\\n\
		\signature %s =\n\
		\  sig\n\
	      \", [fmtData sigid]);
	    prSpecGrps (2, specs);
	    pr "  end;\n";
	    TextIO.closeOut outS
	  end

    fun genSigFile (codeDir, cpys, iter) = (case (findSigBody iter)
	   of NONE => NONE
	    | (SOME({file, sigId, sigBody}, iter')) => let
		val specs = MLSpec.getSigbodyContents sigBody
		val file = OS.Path.joinDirFile{dir=codeDir, file=file}
		in
		  generateFile {file=file, cpys=cpys, sigid=sigId, specs=specs};
		  SOME iter'
		end
	  (* end case *))

    fun doInputFile config = let
	  val outDir = Option.valOf(MLDocConfig.getStr(config, ["OutDir"]))
	  val parse = RunMLDocSGMLS.parseFile {config=config, includeEntities=[]}
	  val basename = RunMLDocSGMLS.basename config
	  fun doFile inFile = let
		val {dir, ...} = basename inFile
		val outDir = OS.Path.concat(outDir, dir)
		val doc = parse inFile
		val iter = M.iterateList doc
		val (cpys, iter) = (case (getCopyrights iter)
		       of (SOME(cpys, iter)) => (cpys, iter)
			| NONE => (
			    Error.warning("empty input file", []);
			    ([], iter))
		      (* end case *))
		fun doSigs iter = (case genSigFile (outDir, cpys, iter)
		       of NONE => ()
			| (SOME iter') => doSigs iter'
		      (* end case *))
		in
		  doSigs iter
		end
		  handle (Fail msg) =>
		    Error.error (F.format "Error in \"%s\": %s\n" [
			F.STR inFile, F.STR msg
		      ])
	  in
	    doFile
	  end

    val options = StdOptions.stdOptions @ [
	    StdOptions.outDir "Sigs"
	  ]

    fun doOptions args = let
	  val {config, otherArgs} = C.configure {
		  appl = "Extract-Sig",
		  opts = options,
		  args = args
		}
	  in
	    {config = config, files = otherArgs}
	  end

    fun main (cmdName, cmdLine) = let
	  val {config, files} = doOptions cmdLine
	  in
	    List.app (doInputFile config) files;
	    OS.Process.success
	  end
	    handle e => (Error.uncaughtExn e; OS.Process.failure)

 end
