(* hrefs.sml
 *
 * COPYRIGHT (c) 1996 AT&T Research.
 * COPYRIGHT (c) 1999 Bell Labs, Lucent Technologies.
 *
 * Support for hypertext references.  The format for the generated URL for an
 * ML identifier is:
 *
 *	[file "#"] [ context-kind ":" ] qualified-id ":" kind [":SPEC"]
 *
 * where
 *
 *	context-kind  ::=  "SIG" | "STR" | "FCT" | "ARG"
 *	kind          ::=  "SIG" | "STR" | "FCT" | "SHARE" | "EXN" | "TY" | "VAL"
 *
 * The file name prefix is for non-local references only.  The context-kind
 * specifies the kind of the binding context (top-level, signature, etc.);
 * top-level identifiers have no context-kind prefix.  The qualified-id is
 * the full name of the identifier, and the kind is its class (signature,
 * value, etc.).  The optional "SPEC" as appended to the specification site
 * URL, while the other name is used for the definition site.  Constructor
 * IDs get mapped to their owning datatype SPEC URL for specification sites,
 * and to a VAL URL at their description site.
 *
 * TODO:
 *  check validity of local references
 *)

structure HRefs : sig

    datatype xref_kind
      = SigRef | StrRef | FctRef | ShareRef | TyRef | ExnRef | ConRef | ValRef

    val mkRemoteRef : MLDocIndex.index -> {
	    base : string option,
	    bindCtx : MLDocIndex.binding_context,
	    id : string, kind : xref_kind
	  } -> HTML.text -> HTML.text

    val xrefToHTML : HTMLContext.context -> {
	    xref : MLDocElem.element, id : string
	  } -> (HTML.text -> HTML.text)
	(* convert an %XREF element to an HTML <A> element *)

    val mkLocalRef : HTMLContext.context -> {
	    kind : xref_kind, localId : string
	  } -> {
	    xref : string -> HTML.text -> HTML.text,
	    anchor : HTML.text -> HTML.text
	  }
	(* this makes a xref/anchor pair for a definition site URL.
	 * The xref <A> element has the SPEC name, and references
	 * the anchor <A> element.  The localId string is used to form the
	 * anchor and reference URL, while the first argument to the xref
	 * function is the SPEC ID.
	 *)

    val mkSpecAnchor : {
	    bindCtx : MLDocIndex.binding_context, id : string, kind : xref_kind
	  } -> (HTML.text -> HTML.text)
	  (* make an anchor element for a specification site *)

  (* Anchors and hyper-links for sections *)
    val mkSectAnchor : (string option * int * HTML.text) -> HTML.text
    val mkSectRef : (string * string option * int * HTML.text) -> HTML.text

    val makeHRef : string -> (HTML.text -> HTML.text)
	  (* wrap an HTML hyper-link around some text *)
    val mkAnchor : string -> (HTML.text -> HTML.text)
	  (* wrap an HTML anchor element around some text *)

  end = struct

    structure C = HTMLContext
    structure M = MLDocMarkup
    structure E = MLDocElem
    structure I = MLDocIndex
    structure F = Format

    datatype xref_kind
      = SigRef | StrRef | FctRef | ShareRef | TyRef | ExnRef | ConRef | ValRef

  (* extract the file name from a file index entry *)
    fun fileName fileInfo = Atom.toString(I.File.name fileInfo)

  (* split a qualified ID into a list of strings *)
    val splitId = String.fields (fn #"." => true | _ => false)

    fun kindSuffix SigRef   = ":SIG"
      | kindSuffix StrRef   = ":STR"
      | kindSuffix FctRef   = ":FCT"
      | kindSuffix ShareRef = ":SHARE"
      | kindSuffix TyRef    = ":TY"
      | kindSuffix ExnRef   = ":EXN"
      | kindSuffix ConRef   = ":TY"
      | kindSuffix ValRef   = ":VAL"

    fun kindName SigRef   = "signature"
      | kindName StrRef   = "structure"
      | kindName FctRef   = "functor"
      | kindName TyRef    = "type"
      | kindName ShareRef = "sharing"
      | kindName ExnRef   = "exception"
      | kindName ConRef   = "constructor"
      | kindName ValRef   = "value"

    fun contextToString bindCtx = let
	  fun joinPath [id] = [Atom.toString id, "."]
	    | joinPath (id::r) = (Atom.toString id) :: "." :: joinPath r
	  val strs = (case bindCtx
		 of I.TOPbound => []
		  | (I.SIGbound p) => "SIG:" :: joinPath p
		  | (I.STRbound p) => "STR:" :: joinPath p
		  | (I.FCTbound p) => "FCT:" :: joinPath p
		  | (I.ARGbound p) => "ARG:" :: joinPath p
		(* end case *))
	  in
	    String.concat strs
	  end

    fun mlTag (bindCtx, id, kind) = F.format "%s%s%s" [
	    F.STR(contextToString bindCtx),
	    F.STR id,
	    F.STR(kindSuffix kind)
	  ]

    fun makeHRef tag htext = MakeHTML.mkA_HREF{href=tag, content=htext}

  (* Given a binding context and a (possibly qualified) ID, return an extended
   * context and unqualified ID.
   *)
    fun qualifyId (bindCtx, name) = (case List.rev(splitId name)
	   of [id] => (bindCtx, id)
	    | (id::qid) => let
		fun error () = (
		      Error.warning(
			"binding context \"%s\" and qualified ID %s disagree",
			[F.STR(contextToString bindCtx), F.STR name]);
		      (bindCtx, id))
		fun strip bindCons (ctxPath, qidPath) = let
		      fun stripPrefix (a::r1, b::r2) =
			    if (Atom.toString a = b)
			      then match (r1, r2)
			      else stripPrefix (r1, b::r2)
			| stripPrefix args = match args
		      and match ([], []) = (bindCtx, id)
			| match ([], r2) =
			    (bindCons(ctxPath @ (map Atom.atom r2)), id)
			| match (a::r1, b::r2) = if (Atom.toString a = b)
			    then match (r1, r2)
			    else error()
			| match _ = error()
		      in
			stripPrefix (ctxPath, qidPath)
		      end
(***
val strip = fn cons => fn (path, qid) => (
print "qualifyId.strip: path = ";
print(ListFormat.fmt{init="[", final="]", fmt=Atom.toString, sep=";"} path);
print ", qid = ";
print(ListFormat.fmt{init="[", final="]", fmt=(fn x => x), sep=";"} qid);
print "\n";
strip cons (path, qid))
***)
		in
(***
print(concat["qualifyId: name = ", name, ", id = ", id, "\n"]);
***)
		  case bindCtx
		   of I.TOPbound => error()
		    | I.SIGbound path => strip I.SIGbound (path, List.rev qid)
		    | I.STRbound path => strip I.STRbound (path, List.rev qid)
		    | I.FCTbound path => strip I.FCTbound (path, List.rev qid)
		  (* end case *)
		end
	  (* end case *))

  (* make a reference to a remote URL.  This will be a SPEC URL. *)
    fun mkRemoteRef index {base, bindCtx, id, kind} = let
	  fun mkLink (NONE, _, _, _) = (
		Error.warning("unable to find %s \"%s%s\"", [
		    F.STR(kindName kind), F.STR(contextToString bindCtx),
		    F.STR id
		  ]);
		(fn htext => htext))
	    | mkLink (SOME entry, getFile, getCtx, getId) = (case base
		 of NONE => makeHRef(F.format "%s.html#%s:SPEC" [
			F.STR(fileName(getFile entry)),
			F.STR(mlTag (getCtx entry, getId entry, kind))
		      ])
		  | (SOME b) => makeHRef(F.format "http://%s/%s.html#%s:SPEC" [
			F.STR b, F.STR(fileName(getFile entry)),
			F.STR(mlTag (getCtx entry, getId entry, kind))
		      ])
		(* end case *))
	  val name = Atom.atom id
	  fun id' _ = id
	  fun top _ = I.TOPbound
	  in
	    case (bindCtx, kind)
	     of (I.TOPbound, SigRef) =>
		  mkLink (I.findSig(index, name), I.Sig.file, top, id')
	      | (I.TOPbound, StrRef) =>
		  mkLink (I.findStr(index, name), I.Str.file, top, id')
	      | (I.TOPbound, FctRef) =>
		  mkLink (I.findFct(index, name), I.Fct.file, top, id')
	      | (_, StrRef) => mkLink (
		    I.findSubstr(index, (bindCtx, name)),
		    I.Str.file,
		    I.Str.context,
		    id')
	      | (_, TyRef) => mkLink (
		    I.findType(index, (bindCtx, name)),
		    I.Ty.file,
		    I.Ty.context,
		    id')
	      | (_, ExnRef) => mkLink (
		    I.findExn(index, (bindCtx, name)),
		    I.Exn.file,
		    I.Exn.context,
		    id')
	      | (_, ConRef) => mkLink (
		    I.findCon(index, (bindCtx, name)),
		    I.Ty.file,
		    I.Ty.context,
		    Atom.toString o I.Ty.name)
	      | (_, ValRef) => mkLink (
		    I.findVal(index, (bindCtx, name)),
		    I.Val.file,
		    I.Val.context,
		    id')
	      | _ => raise Fail "mkRemoteRef: bogus context/ref"
	    (* end case *)
	  end
 
    fun xrefToHTML ctx {xref, id} = let
	  fun nop htext = htext
	  fun homeToBindCtx NONE = C.bindingContext ctx  (* local binding *)
	    | homeToBindCtx (SOME modId) = let
		fun doId s = List.map Atom.atom (splitId s)
		in
		  case modId
		   of E.TOPID => I.TOPbound
		    | (E.SIGID s) => I.SIGbound(doId s)
		    | (E.STRID s) => I.STRbound(doId s)
		    | (E.FCTID s) => I.FCTbound(doId s)
		  (* end case *)
		end
	  fun mkXRef (SOME docName, bindCtx, kind) = let
	      (* strip the \| |\ that marks SDATA *)
		val docName = String.substring(docName, 2, size docName - 4)
		in
		  case C.docIndex(ctx, docName)
		   of (SOME info) => let
			val base = valOf(C.docBaseURL(ctx, docName))
		        val (bindCtx', id') = qualifyId(bindCtx, id)
		        in
			  mkRemoteRef info {
			      base=SOME base, bindCtx=bindCtx', id=id', kind=kind
			    }
		        end
		    | NONE => (
		        Error.warning(
			  "reference to \"%s\" ignored", [F.STR docName]);
		        nop)
		  (* end case *)
		end
	    | mkXRef (NONE, bindCtx, kind) = let
		val (bindCtx', id') = qualifyId(bindCtx, id)
		in
		  mkRemoteRef (C.index ctx) {
		      base=NONE, bindCtx=bindCtx', id=id', kind=kind
		    }
		end
	  in
	    case xref
	     of E.SIGREF{document, nolink=false, ...} =>
		  mkXRef (document, I.TOPbound, SigRef)
	      | E.STRREF{document, home, nolink=false, ...} =>
		  mkXRef (document, homeToBindCtx home, StrRef)
	      | E.FCTREF{document, nolink=false, ...} =>
		  mkXRef (document, I.TOPbound, FctRef)
	      | E.TYREF{document, home, nolink=false, ...} =>
		  mkXRef (document, homeToBindCtx home, TyRef)
	      | E.EXNREF{document, home, nolink=false, ...} =>
		  mkXRef (document, homeToBindCtx home, ExnRef)
	      | E.CONREF{document, home, nolink=false, ...} =>
		  mkXRef (document, homeToBindCtx home, ConRef)
	      | E.VALREF{document, home, nolink=false, ...} =>
		  mkXRef (document, homeToBindCtx home, ValRef)
	      | E.AREF{document=NONE, tag} => (case C.findAnchor (ctx, tag)
		   of (SOME anchorEntry) => makeHRef (F.format "%s.html#%s" [
			  F.STR(fileName(I.Anchor.file anchorEntry)), F.STR tag
			])
		    | NONE => (
			Error.warning ("unable to locate anchor \"%s\"",
			  [F.STR tag]);
			nop)
		  (* end case *))
	      | (E.AREF _) => (
		  Error.warning ("external documents not yet supported!", []);
		  nop)
	      | _ => nop
	    (* end case *)
	  end

  (* wrap an HTML anchor element around some text *)
    fun mkAnchor tag htext = MakeHTML.mkA_NAME {name=tag, content=htext}

  (* make an anchor element for a specification site *)
    fun mkSpecAnchor {bindCtx, id, kind} = let
	  val tag = F.format "%s:SPEC" [F.STR(mlTag (bindCtx, id, kind))]
	  in
	    mkAnchor tag
	  end

  (* this makes a xref/anchor pair for a definition site URL.
   * The xref <A> element has the SPEC name, and references
   * the anchor <A> element.
   *)
    fun mkLocalRef ctx {kind, localId} = let
	  val bindCtx = C.bindingContext ctx
	  val coreTag = mlTag (bindCtx, localId, kind)
	  val xrefTag = F.format "#%s" [F.STR coreTag]
	  fun mkXRef specId htext = HTML.A{
		  title = NONE, 
		  href = SOME xrefTag,
		  name =
		    SOME(F.format "%s:SPEC" [F.STR(mlTag (bindCtx, specId, kind))]),
		  rel = NONE, rev = NONE,
		  content = htext
		}
	  in
	    { xref = mkXRef,
	      anchor = mkAnchor coreTag
	    }
	  end

    fun mkSectAnchor (SOME label, id, text) = mkAnchor label text
      | mkSectAnchor (NONE, id, text) = mkAnchor ("section:" ^ Int.toString id) text

    fun mkSectRef (file, SOME label, id, text) =
	  makeHRef (concat[file, ".html#", label]) text
      | mkSectRef (file, NONE, id, text) =
	  makeHRef (concat[file, ".html#section:", Int.toString id]) text

  end
