(* file-tree.sml
 *
 * COPYRIGHT (c) 1996 AT&T Research
 *
 * Code for creating the inclusion forest of files from an index.
 *)

structure FileTree : sig

    datatype file_nd = FILE of {
	file : MLDocIndex.File.entry,
	parent : MLDocIndex.File.entry option,
	children : file_nd list
      }

    val mkForest : MLDocIndex.index -> {
	    trees : file_nd list,
	    nodeMap : MLDocIndex.File.entry -> file_nd
	  }

  end = struct

    structure IR = IndexRep

    datatype file_nd = FILE of {
	file : MLDocIndex.File.entry,
	parent : MLDocIndex.File.entry option,
	children : file_nd list
      }

  (* allocate a table for mapping file names to a pair of their file
   * entry and their parent reference.  We construct this table in two passes:
   * the first inserts a node for eachfile, and the second sets the parent
   * links.
   *)
    fun mkParentLinkTbl fileTbl = let
	  val tbl = AtomTable.mkTable (
		AtomTable.numItems fileTbl,
		Fail "parentLinkTbl")
	  fun ins1 (name, entry) = AtomTable.insert tbl (name, (entry, ref NONE))
	  fun ins2 (entry as IR.FILEentry{incl, ...}) = let
		fun ins name = (case AtomTable.find tbl name
		       of SOME(_, parent as ref NONE) => parent := SOME entry
			| SOME(_, ref(SOME _)) => Error.error(
			    Format.format "Not a tree: %s has multuple parents"
			      [Format.ATOM name])
			| NONE => ()
		      (* end case *))
		in
		  List.app ins (! incl)
		end
	  in
	    AtomTable.appi ins1 fileTbl;
	    AtomTable.app ins2 fileTbl;
	    tbl
	  end

    fun mkForest (IR.IDX{fileTbl, ...}) = let
	  val parentTbl = mkParentLinkTbl fileTbl
	  val mapTbl = AtomTable.mkTable (AtomTable.numItems fileTbl, Fail "mapTbl")
	  val roots = ref []
	  fun buildTree fileName = (case AtomTable.find mapTbl fileName
		 of (SOME nd) => nd
		  | NONE => let
		      val (entry as IR.FILEentry{incl, ...}, ref parent) =
			    AtomTable.lookup parentTbl fileName
		      fun filterFile n = (case AtomTable.find parentTbl n
			     of NONE => false
			      | (SOME _) => true
			    (* end case *))
		      val incl' = List.filter filterFile (! incl)
		      val nd = FILE{
			      file = entry,
			      parent = parent,
			      children = List.map buildTree incl'
			    }
		      in
			AtomTable.insert mapTbl (fileName, nd);
			if isSome parent then () else roots := nd :: !roots;
			nd
		      end
		(* end case *))
	  in
	    AtomTable.appi (ignore o buildTree o #1) parentTbl;
	    { trees = !roots,
	      nodeMap = fn (IR.FILEentry{name, ...}) => AtomTable.lookup mapTbl name
	    }
	  end

  end;
