(* ppspec-fn.sml
 *
 * COPYRIGHT (c) 1996 AT&T Research.
 *
 * This is a pretty-printer for SML specifications (in signatures).  It is
 * mainly concerned with printing types in an attractive way.
 *)

functor PPSpecFn (structure Dst : sig

    type stream
    type id
    val idSize : id -> int
    val putId : (stream * id) -> stream
    val putc : (stream * char) -> stream
    val puts : (stream * string) -> stream
    val newline : stream -> stream
    val indent : (stream * int) -> stream

  end) : sig

    type stream
    type id

    type region = {indent : int, width : int}
	(* a region is (width-indent) characters wide *)

(**
    val ppTySpec : region -> (stream * ?) -> stream
    val ppExnSpec : region -> (stream * ?) -> stream
    val ppDTSpec : region -> (stream * ?) -> stream
    val ppValSpec : region -> (stream * ?) -> stream
**)

    val ppType : region -> (stream * int * id MLDocType.ty) -> stream

  end = struct

    structure Ty = MLDocType

    type stream = Dst.stream
    type id = Dst.id

    type region = {indent : int, width : int}

    datatype sized_ty
      = VarTy of (int * string)
      | BaseTy of (int * sized_ty list * id)
      | FnTy of (int * sized_ty list)
      | TupleTy of (int * sized_ty list)
      | RecordTy of (int * (string * sized_ty) list)
      | ParenTy of (int * sized_ty)

    fun sizeOf (VarTy(n, _)) = n
      | sizeOf (BaseTy(n, _, _)) = n
      | sizeOf (FnTy(n, _)) = n
      | sizeOf (TupleTy(n, _)) = n
      | sizeOf (RecordTy(n, _)) = n
      | sizeOf (ParenTy(n, _)) = n

    fun computeSize ty = let
	  fun sz (Ty.VarTy s) = (size s, VarTy(size s, s))
	    | sz (Ty.BaseTy([], id)) = let
		val n = Dst.idSize id
		in
		  (n, BaseTy(n, [], id))
		end
	    | sz (Ty.BaseTy([ty], id)) = let
		val n = Dst.idSize id
		val (n', ty) = sz ty
		in
		  (n'+n+1, BaseTy(n'+n+1, [ty], id))
		end
	    | sz (Ty.BaseTy(tys, id)) = let
		val n = Dst.idSize id
		fun f (ty, (n, tys)) = let
		      val (n', ty) = sz ty
		      in
			(n + n' + 2, ty :: tys)
		      end
		val (wid, tys) = List.foldr f (n+1, []) tys
		in
		  (wid, BaseTy(wid, tys, id))
		end
	    | sz (Ty.FnTy(ty1, ty2)) = let
		val (n1, ty1) = sz ty1
		val (n2, ty2) = sz ty2
		val wid = n1+n2+4
		in
		  case ty2
		   of (FnTy(_, tys)) => (wid, FnTy(wid, ty1::tys))
		    | _ => (wid, FnTy(wid, [ty1, ty2]))
		  (* end case *)
		end
	    | sz (Ty.TupleTy tys) = let
		fun f (ty, (n, tys)) = let
		      val (n', ty) = sz ty
		      in
			(n + n' + 3, ty :: tys)
		      end
		val (wid, tys) = List.foldr f (~3, []) tys
		in
		  (wid, TupleTy(wid, tys))
		end
	    | sz (Ty.RecordTy[]) = (2, RecordTy(2, []))
	    | sz (Ty.RecordTy fields) = let
		fun f ((lab, ty), (n, fields)) = let
		      val (n', ty) = sz ty
		      in
		      (* include space for " : ", and ", " *)
			(n + size lab + n' + 5, (lab, ty) :: fields)
		      end
		val (wid, fields) = List.foldr f (0, []) fields
		in
		  (wid, RecordTy(wid, fields))
		end
	    | sz (Ty.ParenTy ty) = let val (n, ty) = sz ty
		in
		  (n+2, ParenTy(n+2, ty))
		end
	  in
	    sz ty
	  end 

    fun prList (prItem, prSep) (strm, items) = let
	  fun pr (strm, []) = strm
	    | pr (strm, [x]) = prItem(strm, x)
	    | pr (strm, x::r) = pr (prSep (prItem (strm, x)), r)
	  in
	    pr (strm, items)
	  end

    fun prSep sepStr strm = Dst.puts(strm, sepStr)

  (* print a type to the stream on a single line *)
    fun prTy (strm, VarTy(_, s)) = Dst.puts(strm, s)
      | prTy (strm, BaseTy(_, [], id)) = Dst.putId(strm, id)
      | prTy (strm, BaseTy(_, [ty], id)) =
	  Dst.putId (Dst.putc (prTy (strm, ty), #" "), id)
      | prTy (strm, BaseTy(_, tys, id)) = let
	  val strm = Dst.putc (strm, #"(")
	  val strm = prList (prTy, prSep ", ") (strm, tys)
	  val strm = Dst.puts (strm, ") ")
	  val strm = Dst.putId (strm, id)
	  in
	    strm
	  end
      | prTy (strm, FnTy(_, tys)) =
	  prList (prTy, prSep " -> ") (strm, tys)
      | prTy (strm, TupleTy(_, tys)) =
	  prList (prTy, prSep " * ") (strm, tys)
      | prTy (strm, RecordTy(_, fields)) = let
	  val strm = Dst.putc (strm, #"{")
	  val strm = prList (prField, prSep ", ") (strm, fields)
	  val strm = Dst.putc (strm, #"}")
	  in
	    strm
	  end
      | prTy (strm, ParenTy(_, ty)) =
	  Dst.putc (prTy (Dst.putc (strm, #"("), ty), #")")

    and prField (strm, (lab, ty)) =
	  prTy (Dst.puts (Dst.puts (strm, lab), " : "), ty)

    fun fits (ty, availSz) = (availSz >= sizeOf ty)

  (* compute a layout for the type; this is represented as a
   * function from strm to strm.
   *)
    fun layout {indent, width, initPos, ty} = let
	  val avail = width-indent
	  fun oneLn (ty, false) strm = prTy(strm, ty)
	    | oneLn (ty, true) strm = Dst.newline(prTy(strm, ty))
	  fun f (FnTy(_, tys)) =
		if (List.all (fn ty => fits(ty, avail-3))) tys
		  then let
		    fun prSep strm =
			  Dst.puts (
			    Dst.indent (Dst.newline strm, indent),
			    "-> ")
		    val prTys = prList (prTy, prSep)
		    in
		      if (fits(List.hd tys, avail-initPos))
			then fn strm => prTys(strm, tys)
			else fn strm => prTys(Dst.newline strm, tys)
		    end
		  else raise Fail "unimplemented"
	    | f (RecordTy(_, fields)) = let
		fun prSep strm =
		      Dst.indent (Dst.newline (Dst.putc (strm, #",")),
		      indent+4)
		val prFields = prList (prField, prSep)
		in
		  fn strm => let
		    val strm = Dst.putc(strm, #"{")
		    val strm = Dst.indent(Dst.newline strm, indent+4)
		    val strm = prFields (strm, fields)
		    val strm = Dst.indent(Dst.newline strm, indent+2)
		    val strm = Dst.putc(strm, #"}")
		    in
		      strm
		    end
		end
	    | f (TupleTy(_, tys)) = let
		fun prSep strm =
		      Dst.puts (
			Dst.indent (Dst.newline strm, indent+4),
			"* ")
		val prItems = prList (prTy, prSep)
		in
		  fn strm => let
		    val strm = Dst.putc(strm, #"(")
		    val strm = Dst.indent(Dst.newline strm, indent+4)
		    val strm = prItems (strm, tys)
		    val strm = Dst.indent(Dst.newline strm, indent+2)
		    val strm = Dst.putc(strm, #")")
		    in
		      strm
		    end
		end
	    | f _ = raise Fail "unable to prettyprint large atomic type"
	  in
	    if (fits(ty, avail))
	      then fn strm => prTy(Dst.indent(Dst.newline strm, indent), ty)
	      else f ty
	  end

    fun ppType {indent, width} (strm, initPos, ty) = let
	  val (n, ty) = computeSize ty
	  in
	    if (n <= width-(indent+initPos))
	      then prTy (strm, ty)
	      else layout {
		  indent=indent, width=width, initPos=initPos, ty=ty
		} strm
	  end

  end
